"""Config loader."""

import datetime
import importlib
import json
import os
import sys
from glob import glob

import jinja2
from jinja2 import meta

import opentaskpy.otflogging
from opentaskpy.exceptions import (
    DuplicateConfigFileError,
    VariableResolutionTooDeepError,
)

MAX_DEPTH = 5


class ConfigLoader:
    """Class responsible for loading and validating config files."""

    def __init__(self, config_dir: str) -> None:
        """Parse config files and expand variables.

        Args:
            config_dir (str): The path to the config files to load.
        """
        self.logger = opentaskpy.otflogging.init_logging(__name__)
        self.config_dir = config_dir
        self.global_variables: dict = {}

        self.logger.log(12, f"Looking in {self.config_dir}")

        # Force Jinja2 to log undefined variables
        jinja2.make_logging_undefined(logger=self.logger, base=jinja2.Undefined)
        self.template_env = jinja2.Environment(undefined=jinja2.StrictUndefined)
        self.template_env.filters["delta_days"] = self.delta_days
        self.template_env.filters["delta_hours"] = self.delta_hours

        self._load_global_variables()

        self.lazy_load = False
        if os.environ.get("OTF_LAZY_LOAD_VARIABLES", None) == "1":
            self.lazy_load = True

        self._resolve_templated_variables(lazy_load=self.lazy_load)

    def get_global_variables(self) -> dict:
        """Return the set of global variables that have been assigned via config files.

        This will first check for anything that has been overridden in the environment
        first, and replace those if necessary.

        Returns:
            dict: _description_
        """
        # Before we return the variables, we need to check for any environment variables that match the same name, and are set
        # if this is the case, then we replace the value with the environment variable
        for key, _ in self.global_variables.items():
            if key in os.environ:
                self.logger.info(
                    f"Overriding global variable ({key}: {self.global_variables[key]})"
                    f" with environment variable ({os.environ[key]})"
                )
                self.global_variables[key] = os.environ[key]

        return self.global_variables

    def delta_days(self, value: datetime.datetime, days: int) -> datetime.datetime:
        """Returns a new datetime object + or - the number of delta days.

        Args:
            value (datetime.datetime): Starting datetime object
            days (int): Days to increment or decrement the value

        Returns:
            datetime.datetime: New datetime object with the delta applied
        """
        return value + datetime.timedelta(days)

    def delta_hours(self, value: datetime.datetime, hours: int) -> datetime.datetime:
        """Returns a new datetime object + or - the number of delta hours.

        Args:
            value (datetime.datetime): Starting datetime object
            hours (int): Hours to increment or decrement the value

        Returns:
            datetime.datetime: New datetime object with the delta applied
        """
        return value + datetime.timedelta(hours=hours)

    def template_lookup(self, plugin: str, **kwargs) -> str:  # type: ignore[no-untyped-def]
        """Lookup function used by Jinja.

        This function is responsible for calling the lookup plugins to evaluate custom
        variables. This imports the necessary Python module and calls the lookup
        function's run method.

        Args:
            plugin (str): The name of the plugin to call
            **kwargs: The arguments to pass to the plugin

        Returns:
            str: The result of the lookup plugin
        """
        self.logger.log(
            11, f"Got call to lookup function {plugin} with kwargs {kwargs}"
        )

        # Append the globals to the kwargs
        kwargs["globals"] = self.global_variables

        # Import the plugin if its not already loaded
        if f"opentaskpy.plugins.lookup.{plugin}" not in sys.modules:
            # Check the module is loadable
            try:
                importlib.import_module(f"opentaskpy.plugins.lookup.{plugin}")
            except ModuleNotFoundError:
                self.logger.log(
                    11,
                    (
                        f"Module not found: opentaskpy.plugins.lookup.{plugin}. Looking"
                        " in plugins directory instead"
                    ),
                )

        # If we haven't loaded the plugin yet, then look in the cfg/plugins directory and see if we can find it
        if f"opentaskpy.plugins.lookup.{plugin}" not in sys.modules:
            plugin_path = f"{self.config_dir}/plugins/lookup/{plugin}.py"
            if os.path.isfile(plugin_path):
                spec = importlib.util.spec_from_file_location(  # type: ignore[attr-defined]
                    f"opentaskpy.plugins.lookup.{plugin}", plugin_path
                )
                module = importlib.util.module_from_spec(spec)  # type: ignore[attr-defined]
                spec.loader.exec_module(module)
                sys.modules[f"opentaskpy.plugins.lookup.{plugin}"] = module

        # If we are in noop mode, then don't actually run the plugin
        if "OTF_NOOP" in os.environ and os.environ["OTF_NOOP"] == "true":
            return "noop"

        # Run the run function of the imported module
        return str(
            getattr(  # noqa: B009
                sys.modules[f"opentaskpy.plugins.lookup.{plugin}"], "run"
            )(**kwargs)
        )

    # TASK DEFINITION FIND FILE
    def load_task_definition(self, task_id: str) -> dict:
        """Load the task definition from the config directory.

        Args:
            task_id (str): The id of the task to load

        Raises:
            DuplicateConfigFileError: Raised if more than one config file is found
            matching the task_id
            FileNotFoundError: Raised if no config file is found matching the task_id

        Returns:
            dict: A dictionary representing the task definition
        """
        json_config = glob(f"{self.config_dir}/**/{task_id}.json", recursive=True)
        json_config.extend(
            glob(f"{self.config_dir}/**/{task_id}.json.j2", recursive=True)
        )

        if not json_config or len(json_config) != 1:
            if len(json_config) > 1:
                raise DuplicateConfigFileError(
                    f"Found more than one task with name: {task_id}"
                )

            raise FileNotFoundError(f"Couldn't find task with name: {task_id}")

        found_file = json_config[0]
        self.logger.log(12, f"Found: {found_file}")

        task_definition = self._enrich_variables(found_file)

        # If the task has any variables, we need to check if they're overridden by environment variables
        if "variables" in task_definition:
            for key, _ in task_definition["variables"].items():
                if key in os.environ:
                    task_definition["variables"][key] = os.environ[key]

        # Check to see if this is not a batch type
        if "type" in task_definition and task_definition["type"] == "batch":
            self.logger.debug("Cannot apply overrides to batch tasks")
            return task_definition

        # Finally, attributes of the task definition can also be overridden by environment variables
        # e.g. OTF_OVERRIDE_TRANSFER_SOURCE_HOSTNAME will override ["source"]["hostname"], we need to handle this
        # The format is OTF_OVERRIDE_<TASK_TYPE>_<ATTRIBUTE>_<ATTRIBUTE>_<ATTRIBUTE>
        # e.g. OTF_OVERRIDE_TRANSFER_SOURCE_HOSTNAME
        # Some special cases need to be handled, e.g. OTF_OVERRIDE_EXECUTION_PROTOCOL!!SOME_ATTRIBUTE
        # This is where the attribute contains the delimiter itself, so a split by _ will not work
        for key, value in os.environ.items():
            if key.startswith("OTF_OVERRIDE_"):
                # Split the key by _ but handle !! special case first
                if "!!" in key:
                    # Split into parts before and after !!
                    prefix, property_name = key.split("!!")
                    # Split the prefix normally by _ and remove OTF_OVERRIDE_
                    split_key = prefix.split("_")[3:] + [property_name]
                else:
                    # Original behavior
                    split_key = key.split("_")[3:]

                self._apply_env_var_overrides_to_task_definition(
                    task_definition, split_key, value
                )

        return task_definition

    def _apply_env_var_overrides_to_task_definition(
        self, task_definition: dict, split_key: str, value: str
    ) -> None:
        attribute = split_key[0]

        # Match the attribute, to an attribute in the task_definition, bearing in mind that the case may not match
        # e.g. "source" may be "Source"
        for task_definition_attribute in task_definition:
            if task_definition_attribute.lower() == attribute.lower():
                attribute = task_definition_attribute
                break

        # Check that the attribute exists in the definition, if not just ignore it and move on to the next override
        if attribute not in task_definition:
            return

        if len(split_key) == 1:
            self.logger.info(
                f"Overriding {attribute}: {task_definition[attribute]} with {value}"
            )
            task_definition[attribute] = value
        elif isinstance(task_definition[attribute], list):
            # If the current attribute is a list, then we need to get the index from the next element
            # e.g. ["source"]["files"][0]["filename"]
            index = int(split_key[1])
            self._apply_env_var_overrides_to_task_definition(
                task_definition[attribute][index], split_key[2:], value
            )
        else:
            self._apply_env_var_overrides_to_task_definition(
                task_definition[attribute], split_key[1:], value
            )

    # POPULATE VARIABLES INSIDE TASK DEFINITION
    # AND LOAD ADDITIONAL VARIABLES FROM TASK DEFINITION
    def _enrich_variables(self, task_definition_file: str) -> dict:
        active_task_definition = None
        with open(task_definition_file, encoding="utf-8") as json_file:
            json_content = json_file.read()
            template = self.template_env.from_string(json_content)

            # If the file is a Jinja2 template, then we do not allow additional
            # variables to be defined in the task definition.
            # Check the file extension
            if not task_definition_file.endswith(".j2"):
                # From this, convert it to JSON and pull out the variables key if there is one
                task_definition = json.loads(json_content)
                # Extend or replace any local variables for this task
                if "variables" in task_definition:
                    self.global_variables = (
                        self.global_variables | task_definition["variables"]
                    )

                template = self.template_env.from_string(json_content)

            template.globals["utc_now"] = self.now_utc
            template.globals["now"] = self.now_localtime

            # Define lookup function
            template.globals["lookup"] = self.template_lookup

            # Using the active task definition, resolve any templated variables
            # Find any variables that are templated from the template object
            if self.lazy_load:

                ast = self.template_env.parse(json_content)
                undeclared_variables = meta.find_undeclared_variables(ast)

                self.logger.log(
                    12,
                    f"Found undeclared variables: {undeclared_variables}",
                )

                # For each undeclared variable, resolve it
                for undeclared_variable in undeclared_variables:

                    # Get the value of the variable from global variables, if it exists,
                    # if it doesn't it must be a task variable, so this can be resolved below
                    if undeclared_variable not in self.global_variables:
                        self.logger.log(
                            12,
                            f"Variable {undeclared_variable} not found in global variables, must be a task variable",
                        )
                        continue

                    unresolved_variable = self.global_variables[undeclared_variable]

                    # Check the type of the variable, if it's not a string, then dump it into a JSON string
                    converted_variable = False
                    if not isinstance(unresolved_variable, str):
                        self.logger.debug(
                            "Converted unresolved variable into a RAW string"
                        )
                        unresolved_variable = json.dumps(unresolved_variable)
                        converted_variable = True

                    self.logger.info(f"Resolving variable {undeclared_variable}")
                    evaluated_variable = self._resolve_templated_variables_from_string(
                        unresolved_variable
                    )

                    self.logger.log(
                        12,
                        f"Resolved variable {undeclared_variable}",
                    )

                    # If the variable was not a string, then convert it back to the original type
                    if converted_variable:
                        self.logger.debug("Converting variable back into a JSON object")
                        evaluated_variable = json.loads(evaluated_variable)

                    # Now update the global variables with the resolved value
                    self.global_variables[undeclared_variable] = evaluated_variable

            rendered_template = template.render(self.global_variables)
            active_task_definition = dict(json.loads(rendered_template))
            self.logger.log(
                12,
                f"Evaluated task definition: {opentaskpy.otflogging.redact(json.dumps(active_task_definition))}",
            )

        return active_task_definition

    # READ AND PARSE ACTUAL VARIABLE FILES
    def _load_global_variables(self) -> None:
        global_variables: dict = {}
        variable_configs: list[str] = []

        # See if variables file(s) have been overridden via environment variable
        if "OTF_VARIABLES_FILE" in os.environ:
            new_variables_files = os.environ["OTF_VARIABLES_FILE"].split(",")
            for new_variables_file in new_variables_files:
                self.logger.info(f"Overriding variables file with {new_variables_file}")
                # Validate that the file exists
                if not os.path.isfile(new_variables_file):
                    raise FileNotFoundError(
                        f"Couldn't find variables file: {new_variables_file}"
                    )
                variable_configs.append(new_variables_file)

        else:
            file_types = (".json.j2", ".json")
            for file_type in file_types:
                variable_configs.extend(
                    glob(f"{self.config_dir}/**/variables{file_type}", recursive=True)
                )

            if not variable_configs:
                raise FileNotFoundError(
                    "Couldn't find any variables.(json|json.j2) files under"
                    f" {self.config_dir}"
                )

        for variable_file in variable_configs:
            with open(variable_file, encoding="utf-8") as json_file:
                this_variable_config = json.load(json_file)
                global_variables = global_variables | this_variable_config

        self.global_variables = global_variables

    def now_localtime(self) -> datetime.datetime:
        """Return the current time in the local timezone."""
        return datetime.datetime.now().astimezone()

    def now_utc(self) -> datetime.datetime:
        """Return the current time in UTC."""
        return datetime.datetime.utcnow()

    # RESOLVE ANY VARIABLES THAT USE OTHER VARIABLES IN THE VARIABLE FILES
    def _resolve_templated_variables(self, lazy_load: bool = False) -> None:
        # We need to evaluate the variables themselves, in case there's any recursion
        # Convert the variables to a JSON string which we can process with the jinja2 templater
        current_depth = 0
        previous_render = None

        template = self.global_variables

        variables_template = self.template_env.from_string(json.dumps(template))
        variables_template.globals["utc_now"] = self.now_utc
        variables_template.globals["now"] = self.now_localtime

        # Define lookup function
        variables_template.globals["lookup"] = self.template_lookup

        # Don't do anything if lazy loading is enabled
        if lazy_load:
            return

        evaluated_variables = variables_template.render(template)

        while evaluated_variables != previous_render and current_depth < MAX_DEPTH:
            previous_render = evaluated_variables

            variables_template = self.template_env.from_string(evaluated_variables)
            variables_template.globals["utc_now"] = self.now_utc
            variables_template.globals["now"] = self.now_localtime

            # Define lookup function
            variables_template.globals["lookup"] = self.template_lookup
            evaluated_variables = variables_template.render(
                json.loads(evaluated_variables)
            )

            current_depth += 1
            if current_depth >= MAX_DEPTH:
                self.logger.error(
                    "Reached max depth of recursive template evaluation. Please check"
                    " the task as variable definitions for infinite recursion"
                )
                raise VariableResolutionTooDeepError(
                    "Reached max depth of recursive template evaluation"
                )

        self.global_variables = json.loads(evaluated_variables)

    def _resolve_templated_variables_from_string(self, json_content: str) -> str:
        # This has some code duplication with _resolve_templated_variables, but it's
        # easier to keep them separate for now

        # We need to evaluate the variables themselves, in case there's any recursion
        # Convert the variables to a JSON string which we can process with the jinja2 templater
        current_depth = 0
        previous_render = None

        variables_template = self.template_env.from_string(json_content)
        variables_template.globals["utc_now"] = self.now_utc
        variables_template.globals["now"] = self.now_localtime

        # Define lookup function
        variables_template.globals["lookup"] = self.template_lookup

        evaluated_variables = variables_template.render(self.global_variables)

        while evaluated_variables != previous_render and current_depth < MAX_DEPTH:
            previous_render = evaluated_variables

            variables_template = self.template_env.from_string(evaluated_variables)

            variables_template.globals["utc_now"] = self.now_utc
            variables_template.globals["now"] = self.now_localtime

            # Define lookup function
            variables_template.globals["lookup"] = self.template_lookup
            evaluated_variables = variables_template.render(self.global_variables)

            current_depth += 1
            if current_depth >= MAX_DEPTH:
                self.logger.error(
                    "Reached max depth of recursive template evaluation. Please check"
                    " the task as variable definitions for infinite recursion"
                )
                raise VariableResolutionTooDeepError(
                    "Reached max depth of recursive template evaluation"
                )

        return str(evaluated_variables)



"""Task handler for running transfers."""

import random
import shutil
import threading
import time
from importlib import import_module
from math import ceil, floor
from os import environ, getpid, makedirs, path, remove
from sys import modules
from typing import NamedTuple

import gnupg

import opentaskpy.otflogging
from opentaskpy import exceptions
from opentaskpy.remotehandlers.remotehandler import RemoteTransferHandler
from opentaskpy.taskhandlers.taskhandler import TaskHandler

# Full transfers expect that the remote host has a base install of python3
# We transfer over the wrapper script to the remote host and trigger it, which is responsible
# for doing some of the more complex work, rather than triggering a tonne of shell commands


class DefaultProtocolCharacteristics(NamedTuple):
    """Class defining the configuration for default protocols."""

    module: str
    class_: str


TASK_TYPE = "T"
DEFAULT_PROTOCOL_MAP = {
    "ssh": DefaultProtocolCharacteristics(
        "opentaskpy.remotehandlers.ssh", "SSHTransfer"
    ),
    "sftp": DefaultProtocolCharacteristics(
        "opentaskpy.remotehandlers.sftp", "SFTPTransfer"
    ),
    "email": DefaultProtocolCharacteristics(
        "opentaskpy.remotehandlers.email", "EmailTransfer"
    ),
    "local": DefaultProtocolCharacteristics(
        "opentaskpy.remotehandlers.local", "LocalTransfer"
    ),
    "dummy": DefaultProtocolCharacteristics(
        "opentaskpy.remotehandlers.dummy", "DummyTransfer"
    ),
}
DEFAULT_STAGING_DIR_BASE = "/tmp"  # nosec B108


class Transfer(TaskHandler):  # pylint: disable=too-many-instance-attributes
    """Task handler for running transfers."""

    source_remote_handler: RemoteTransferHandler
    dest_remote_handlers: RemoteTransferHandler = None
    source_file_spec: dict
    dest_file_specs: list[dict] | None = None
    overall_result: bool = False

    def __init__(self, global_config: dict, task_id: str, transfer_definition: dict):
        """Create a new transfer task handler.

        Args:
            global_config (dict): Global configuration dictionary
            task_id (str): Task ID
            transfer_definition (dict): Transfer definition
        """
        self.task_id = task_id
        self.transfer_definition = transfer_definition

        # Set up the local staging directory
        # Allow for a custom staging directory to be set via environment variable
        staging_dir_name = f"OTF_STAGING_{getpid()}.{random.randint(0, 1000000)}"
        if "OTF_STAGING_DIR" in environ:
            self.local_staging_dir = f"{environ['OTF_STAGING_DIR']}/{staging_dir_name}"
        else:
            self.local_staging_dir = f"{DEFAULT_STAGING_DIR_BASE}/{staging_dir_name}"

        self.logger = opentaskpy.otflogging.init_logging(
            "opentaskpy.taskhandlers.transfer", self.task_id, TASK_TYPE
        )

        super().__init__(global_config)

    def return_result(
        self,
        status: int,
        message: str | None = None,
        exception: Exception | None = None,
    ) -> bool:
        """Return the result of the task run.

        Args:
            status (int): The status code to return.
            message (str, optional): The message to return. Defaults to None.
            exception (Exception, optional): The exception to return. Defaults to None.

        Returns:
            bool: The result of the task run.
        """
        # Log the exception
        if exception:
            self.logger.exception(exception)

        # Delete the remote connection objects
        if self.source_remote_handler:
            self.logger.info("Closing source connection")
            self.source_remote_handler.tidy()
        if self.dest_remote_handlers:
            for remote_handler in self.dest_remote_handlers:
                self.logger.info(f"Closing dest connection for {remote_handler}")
                remote_handler.tidy()

        # Remove local staging directory if it exists (and this isn't a local transfer)
        if (
            path.exists(self.local_staging_dir)
            and self.source_file_spec["protocol"]["name"] != "dummy"
            and self.local_staging_dir != self.source_file_spec["directory"]
        ):
            self.logger.info(
                f"Removing local staging directory {self.local_staging_dir}"
            )
            shutil.rmtree(self.local_staging_dir)
        else:
            if path.exists(self.local_staging_dir):
                self.logger.info(
                    "Local staging directory is the same as source directory. Not removing",
                )

        # Call super to do the rest
        return super().return_result(status, message, exception)  # type: ignore[no-any-return]

    def _get_default_class(self, protocol_name: str) -> type:
        class_name = DEFAULT_PROTOCOL_MAP[protocol_name].class_
        module_name = DEFAULT_PROTOCOL_MAP[protocol_name].module

        # Load module
        if module_name not in modules:
            import_module(module_name)

        return getattr(modules[module_name], class_name)  # type: ignore[no-any-return]

    def _set_remote_handlers(self) -> None:
        # Based on the transfer definition, determine what to do first
        self.source_file_spec = self.transfer_definition["source"]
        self.source_file_spec["task_id"] = self.task_id
        source_protocol = self.source_file_spec["protocol"]["name"]

        # Create a list of destination file specs
        if (
            "destination" in self.transfer_definition
            and self.transfer_definition["destination"]
            and isinstance(self.transfer_definition["destination"], list)
        ):
            self.dest_file_specs = self.transfer_definition["destination"]

        # Set the task_id for each file spec
        if self.dest_file_specs:
            for dest_file_spec in self.dest_file_specs:
                dest_file_spec["task_id"] = self.task_id

        # Based on the source protocol pick the appropriate remote handler
        if source_protocol in DEFAULT_PROTOCOL_MAP:
            handler_class = self._get_default_class(source_protocol)
            self.source_remote_handler = handler_class(self.source_file_spec)

        # If not SSH, then it's a non-standard protocol, we need to see if it's loadable
        # load it, and then create the remote handler
        else:
            self.source_remote_handler = super()._get_handler_for_protocol(
                source_protocol, self.source_file_spec
            )

        super()._set_handler_vars(source_protocol, self.source_remote_handler)

        # Based on the destination protocol pick the appropriate remote handler
        if self.dest_file_specs:
            self.dest_remote_handlers = []
            for dest_file_spec in self.dest_file_specs:
                remote_protocol = dest_file_spec["protocol"]["name"]

                # For each host, create a remote handler
                remote_handler = None
                if remote_protocol in DEFAULT_PROTOCOL_MAP:
                    handler_class = self._get_default_class(remote_protocol)
                    remote_handler = handler_class(dest_file_spec)

                else:
                    remote_handler = super()._get_handler_for_protocol(
                        remote_protocol, dest_file_spec
                    )

                self.dest_remote_handlers.append(remote_handler)
                super()._set_handler_vars(remote_protocol, remote_handler)

    def run(self, kill_event: threading.Event | None = None) -> bool:  # noqa: C901
        """Run the transfer task.

        Args:
            kill_event (threading.Event, optional): Event to kill the task. Defaults to None.

        Returns:
            bool: The result of the task run.
        """
        self.logger.info("Running transfer")

        self._set_remote_handlers()

        # If log watching, do that first
        if "logWatch" in self.source_file_spec:
            self.logger.info(
                "Performing a log watch of"
                f" {self.source_file_spec['logWatch']['directory']}/{self.source_file_spec['logWatch']['log']}"
            )

            if self.source_remote_handler.init_logwatch() != 0:
                return self.return_result(
                    1, "Logwatch init failed", exception=exceptions.LogWatchInitError
                )

            timeout_seconds = self.source_file_spec["logWatch"].get("timeout", 60)
            sleep_seconds = self.source_file_spec["logWatch"].get("sleepTime", 10)

            # Now we start the loop to monitor the log file
            start_time = time.time()
            found_log_entry = False
            while floor(time.time() - start_time) <= timeout_seconds:
                if kill_event and kill_event.is_set():
                    return self.return_result(
                        1, "KILLED DUE TO TIMEOUT FROM PARENT BATCH"
                    )

                if self.source_remote_handler.do_logwatch() == 0:
                    found_log_entry = True
                    break

                remaining_seconds = ceil((start_time + timeout_seconds) - time.time())
                if remaining_seconds == 0:
                    break

                # If the sleep time is longer than the time remaining, sleep for that long instead
                if remaining_seconds < sleep_seconds:
                    actual_sleep_seconds = remaining_seconds
                else:
                    actual_sleep_seconds = sleep_seconds

                self.logger.info(
                    f"No entry found in log. Sleeping for {actual_sleep_seconds} secs."
                    f" {remaining_seconds} seconds remain"
                )
                time.sleep(actual_sleep_seconds)

            if found_log_entry:
                self.logger.info("Found pattern in log file")
            else:
                return self.return_result(
                    1,
                    f"No log entry found after {timeout_seconds} seconds",
                    exception=exceptions.LogWatchTimeoutError,
                )

        # If filewatching, do that next
        if "fileWatch" in self.source_file_spec:
            # Setup a loop for the filewatch
            timeout_seconds = self.source_file_spec["fileWatch"].get("timeout", 60)
            sleep_seconds = self.source_file_spec["fileWatch"].get("sleepTime", 10)

            start_time = time.time()
            remote_files: dict = {}

            # Determine if we're doing a plain filewatch, or looking for a different file to what we are transferring
            watch_directory = (
                self.source_file_spec["fileWatch"]["directory"]
                if "directory" in self.source_file_spec["fileWatch"]
                else self.source_file_spec["directory"]
            )
            watch_file_pattern = (
                self.source_file_spec["fileWatch"]["fileRegex"]
                if "fileRegex" in self.source_file_spec["fileWatch"]
                else self.source_file_spec["fileRegex"]
            )
            # Determine if we should check for file conditionals during filewatch, or only after watch is completed
            check_conditionals_during_filewatch = (
                "conditionals" in self.source_file_spec
                and "checkDuringFilewatch" in self.source_file_spec["conditionals"]
            )

            self.logger.info(
                f"Performing a file watch on {watch_directory}/{watch_file_pattern}"
            )

            while (
                not remote_files and floor(time.time() - start_time) <= timeout_seconds
            ):
                if kill_event and kill_event.is_set():
                    return self.return_result(
                        1, "KILLED DUE TO TIMEOUT FROM PARENT BATCH"
                    )

                remote_files = self.source_remote_handler.list_files(
                    directory=watch_directory, file_pattern=watch_file_pattern
                )
                if check_conditionals_during_filewatch and remote_files:
                    self.check_conditionals(remote_files)
                # TODO: #5 Change all references to remote_files to expect a generator # pylint: disable=fixme
                if remote_files:
                    self.logger.info("Filewatch found remote file(s)")
                    break

                remaining_seconds = ceil((start_time + timeout_seconds) - time.time())
                if remaining_seconds == 0:
                    break

                # If the sleep time is longer than the time remaining, sleep for that long instead
                if remaining_seconds < sleep_seconds:
                    actual_sleep_seconds = remaining_seconds
                else:
                    actual_sleep_seconds = sleep_seconds

                # Prevent negative sleep times
                actual_sleep_seconds = max(actual_sleep_seconds, 0)

                self.logger.info(
                    f"No valid files found. Sleeping for {sleep_seconds} secs."
                    f" {remaining_seconds} seconds remain"
                )
                time.sleep(actual_sleep_seconds)

            if (
                remote_files
                and "watchOnly" in self.source_file_spec["fileWatch"]
                and self.source_file_spec["fileWatch"]["watchOnly"]
            ):
                return self.return_result(0, "Just performing filewatch")
            if not remote_files:
                # Only error if error is not set to false
                if (
                    "error" in self.source_file_spec
                    and not self.source_file_spec["error"]
                ):
                    self.logger.info(
                        "Valid files not found after timeout, but error is set to false"
                    )
                    return self.return_result(0, "Valid files not found")

                return self.return_result(
                    1,
                    f"No valid files found after {timeout_seconds} seconds",
                    exception=exceptions.RemoteFileNotFoundError,
                )

        # Determine what needs to be transferred
        source_directory = self.source_file_spec.get("directory", None)
        source_file_pattern = self.source_file_spec.get("fileRegex", None)
        remote_files = self.source_remote_handler.list_files(
            directory=source_directory, file_pattern=source_file_pattern
        )
        decrypted_files = {}
        encrypted_files = {}

        # Loop through the returned files to see if they match any conditionals ( file age, size spec, counts) (if defined)
        if "conditionals" in self.source_file_spec and remote_files:
            remote_files = self.check_conditionals(remote_files)

        if not remote_files and self.source_file_spec["protocol"]["name"] != "dummy":
            if "error" in self.source_file_spec and not self.source_file_spec["error"]:
                return self.return_result(
                    0,
                    "Valid remote files could not be found to transfer. But not erroring"
                    " due to config",
                )

            return self.return_result(
                1,
                "Valid remote files could not be found to transfer",
                exception=exceptions.FilesDoNotMeetConditionsError,
            )

        self.logger.info("Found the following file(s) that match all requirements:")
        for file in remote_files:
            self.logger.info(f" * {file}")

        can_do_encryption = False
        original_file_list = remote_files.copy()
        # If there's a destination file spec, then we need to transfer the files
        if self.dest_file_specs:
            # Loop through all dest_file specs and see if there are any transfers where the source and dest protocols are different
            # If there are, then we need to do a pull transfer first, then a push transfer
            any_different_protocols = False
            i = 0
            for dest_file_spec in self.dest_file_specs:
                different_protocols = False
                if (
                    (
                        "transferType" not in dest_file_spec
                        or dest_file_spec["transferType"] == "push"
                    )
                    # And the destination and source remote handler classes are not the same
                    and (
                        self.source_remote_handler.__class__
                        != self.dest_remote_handlers[i].__class__
                    )
                ):
                    different_protocols = True
                    any_different_protocols = True

                # If there are differences, download the file locally first
                # so it's ready to upload to multiple destinations at once
                if (
                    different_protocols
                    or (
                        "transferType" in dest_file_spec
                        and dest_file_spec["transferType"] == "proxy"
                    )
                    or not self.source_remote_handler.supports_direct_transfer()
                ):
                    # Create local staging dir (if this isn't using the local protocol)
                    if self.source_file_spec["protocol"]["name"] != "local":
                        makedirs(self.local_staging_dir, exist_ok=True)
                    else:
                        self.local_staging_dir = self.source_file_spec["directory"]
                    transfer_result = self.source_remote_handler.pull_files_to_worker(
                        remote_files, self.local_staging_dir
                    )
                    # Since files are being pulled locally, encryption/decryption of the files is possible
                    can_do_encryption = True
                    if transfer_result != 0:
                        return self.return_result(
                            1,
                            "Pull to worker from remote source errored",
                            exception=exceptions.RemoteTransferError,
                        )
                i += 1

            # Before doing any file movements, check to see if file decryption or encryption is
            # required on the source or destination. For any unsupported transferTypes we need to fail here first

            # Supported options are:
            # decrypting when:
            # transferType is proxy, or the protocols are different (in which case it's local anyway)
            # encrypting when:
            # transferType is proxy, or the protocols are different (in which case it's local anyway)

            # Check if decryption is requested
            decryption_requested = (
                "encryption" in self.source_file_spec
                and "decrypt" in self.source_file_spec["encryption"]
                and self.source_file_spec["encryption"]["decrypt"]
            )

            if decryption_requested and not can_do_encryption:
                return self.return_result(
                    1,
                    "Decryption requested but not supported for this transfer",
                    exception=exceptions.DecryptionNotSupportedError,
                )

            # If it's requested and decryption is possible, then we need to decrypt the files
            if decryption_requested and can_do_encryption:
                self.logger.info("Decrypting files")

                # Get the private key from the spec
                private_key = self.source_file_spec["encryption"]["private_key"]

                # For each file in the remote_files list, alter the path to start
                # with the local staging directory instead
                local_files = {}
                for file in remote_files:
                    # Strip the path and replace it with the stating directory
                    local_files[f"{self.local_staging_dir}/{path.basename(file)}"] = (
                        remote_files[file]
                    )

                # Loop through each file and decrypt it using gnupg
                remote_files = self.decrypt_files(local_files, private_key)

                decrypted_files = remote_files.copy()

            i = 0
            for dest_file_spec in self.dest_file_specs:
                encryption_requested = (
                    "encryption" in dest_file_spec
                    and "encrypt" in dest_file_spec["encryption"]
                    and dest_file_spec["encryption"]["encrypt"]
                )
                # Check if encryption is requested
                if encryption_requested and not can_do_encryption:
                    return self.return_result(
                        1,
                        "Encryption requested but not supported for this transfer",
                        exception=exceptions.EncryptionNotSupportedError,
                    )

                # If encryption is requested and its possible, then encrypt the file(s)
                if encryption_requested and can_do_encryption:

                    self.logger.info("Encrypting files")

                    # Get the public key from the spec
                    public_key = dest_file_spec["encryption"]["public_key"]
                    private_key = None

                    if (
                        "sign" in dest_file_spec["encryption"]
                        and dest_file_spec["encryption"]["sign"]
                    ):
                        private_key = dest_file_spec["encryption"]["private_key"]

                    # For each file in the remote_files list, alter the path to start
                    # with the local staging directory instead
                    local_files = {}
                    for file in remote_files:
                        # Strip the path and replace it with the stating directory
                        local_files[
                            f"{self.local_staging_dir}/{path.basename(file)}"
                        ] = remote_files[file]

                    extension = dest_file_spec["encryption"].get(
                        "output_extension", "gpg"
                    )
                    # Loop through each file and encrypt it using gnupg
                    remote_files = self.encrypt_files(
                        local_files, public_key, private_key, extension
                    )

                    encrypted_files = remote_files.copy()

                # Handle the push transfers first
                associated_dest_remote_handler = self.dest_remote_handlers[i]
                # If this is a default push transfer, and both source and dest protocols are the same
                if (
                    (
                        "transferType" not in dest_file_spec
                        or dest_file_spec["transferType"] == "push"
                        # And the destination and source remote handler classes are the same
                    )
                    and not any_different_protocols
                    and self.source_remote_handler.supports_direct_transfer()
                ):
                    transfer_result = self.source_remote_handler.transfer_files(
                        remote_files,
                        dest_file_spec,
                        dest_remote_handler=associated_dest_remote_handler,
                    )
                    if transfer_result != 0:
                        return self.return_result(
                            1,
                            "Remote transfer errored",
                            exception=exceptions.RemoteTransferError,
                        )

                    self.logger.info("Transfer completed successfully")
                # If this is a default push transfer, and source and dest protocols are different
                elif (
                    (
                        "transferType" in dest_file_spec
                        and (
                            dest_file_spec["transferType"] == "push"
                            or dest_file_spec["transferType"] == "proxy"
                        )
                    )
                    or different_protocols
                    or not self.source_remote_handler.supports_direct_transfer()
                ):
                    self.logger.debug(
                        "Transfer protocols are different, or proxy transfer is"
                        " requested"
                    )

                    # For local transfers, the handler needs the list of local files to push
                    # If there was decryption, then the files will be local regardless, so
                    # needs the list of files then too, so it doesn't upload the encrypted files
                    if (
                        self.source_file_spec["protocol"]["name"] != "local"
                        and not decryption_requested
                        and not encryption_requested
                    ):
                        transfer_result = (
                            associated_dest_remote_handler.push_files_from_worker(
                                self.local_staging_dir
                            )
                        )
                    else:
                        transfer_result = (
                            associated_dest_remote_handler.push_files_from_worker(
                                self.local_staging_dir, file_list=remote_files
                            )
                        )

                    if transfer_result != 0:
                        return self.return_result(
                            1,
                            "Push of files to destination errored",
                            exception=exceptions.RemoteTransferError,
                        )

                elif (
                    "transferType" in dest_file_spec
                    and dest_file_spec["transferType"] == "pull"
                ):
                    transfer_result = associated_dest_remote_handler.pull_files(
                        remote_files, self.source_file_spec
                    )
                    if transfer_result != 0:
                        return self.return_result(
                            1,
                            "Remote PULL transfer errored",
                            exception=exceptions.RemoteTransferError,
                        )

                    self.logger.info("Transfer completed successfully")

                # Handle any ownership and permissions changes
                if dest_file_spec["protocol"]["name"] == "ssh":
                    move_result = self.dest_remote_handlers[
                        i
                    ].move_files_to_final_location(remote_files)
                    if move_result != 0:
                        return self.return_result(
                            1,
                            "Error moving file into final location",
                            exception=exceptions.RemoteTransferError,
                        )

                # Create any flag files that might need creating
                if "flags" in dest_file_spec:
                    flag_result = self.dest_remote_handlers[i].create_flag_files()
                    if flag_result != 0:
                        return self.return_result(
                            1,
                            "Error creating flag files",
                            exception=exceptions.RemoteTransferError,
                        )

                i += 1

            if (
                different_protocols
                and self.source_file_spec["protocol"]["name"] != "dummy"
                and self.local_staging_dir != self.source_file_spec["directory"]
            ):
                self.logger.debug("Removing local staging directory")
                if path.exists(self.local_staging_dir):
                    shutil.rmtree(self.local_staging_dir)
        else:
            self.logger.info("Performing filewatch only")

        # If there was encryption done, we need to remove all the encrypted files regardless, and then
        # restore the original remote_files dict
        if encrypted_files.keys():
            for encrypted_file in encrypted_files:
                # Check the file still exists before removing it
                if path.exists(encrypted_file):
                    self.logger.info(f"Removing local encrypted file {encrypted_file}")
                    remove(encrypted_file)

            remote_files = original_file_list

        # Do the same for the decrypted files
        if decrypted_files.keys():
            for decrypted_file in decrypted_files:
                if path.exists(decrypted_file):
                    self.logger.info(f"Removing local decrypted file {decrypted_file}")
                    remove(decrypted_file)

            remote_files = original_file_list

        if "postCopyAction" in self.source_file_spec:
            try:
                self.logger.info("Performing post copy action")
                pca_result = self.source_remote_handler.handle_post_copy_action(
                    remote_files
                )
            except Exception as e:  # pylint: disable=broad-exception-caught
                pca_result = 1
                return self.return_result(
                    1,
                    "Error performing post copy action",
                    exception=e,
                )
            if pca_result != 0:
                return self.return_result(
                    1,
                    "Error performing post copy action",
                    exception=exceptions.RemoteTransferError,
                )

        return self.return_result(0)

    def encrypt_files(
        self,
        files: dict,
        public_key: str,
        private_key: str | None = None,
        extension: str = "gpg",
    ) -> dict:
        """Encrypt files using GPG.

        Args:
            files (dict): Dictionary of files to encrypt.
            public_key (str): Public key to use for encryption.
            private_key (str, optional): Private key to use to sign the files. Defaults to None.
            extension (str, optional): Extension to add to the encrypted files. Defaults to "gpg".

        Returns:
            dict: Dictionary of encrypted files.
        """
        # Use the local staging dir (or the path where the files live) as the gnupg home
        # Get the dirname of the first file in the list
        tmpdir = path.dirname(list(files.keys())[0])

        # Make the gnupg home directory
        makedirs(f"{tmpdir}/.gnupg", exist_ok=True)

        # Set up gnupg
        gpg = gnupg.GPG(gnupghome=f"{tmpdir}/.gnupg")

        # Remove any escaped newline characters from the key
        public_key = public_key.replace("\\n", "\n")

        # Load the public key
        import_result = gpg.import_keys(public_key)

        # Check the key imported OK
        if not import_result.count or import_result.count == 0:
            self.logger.error("Error importing public key")
            raise exceptions.EncryptionError("Error importing public key")

        # Get the fingerprint of the key we just imported
        key_fingerprint = gpg.list_keys()[0]["fingerprint"]
        signing_key = None

        if private_key:
            # Load the private key
            import_result = gpg.import_keys(private_key)

            # Check the key imported OK
            if not import_result.count or import_result.count == 0:
                self.logger.error("Error importing private key")
                raise exceptions.EncryptionError("Error importing private key")

            signing_key = import_result.fingerprints[0]

        encrypted_files = {}
        for file in files:

            # If the filename contains .gpg on the end, then the output file will be just that but without the extension,
            # if it doesn't contain .gpg, then we'll add .gpg to the end
            output_filename = f"{file}.{extension}"

            with open(file, "rb") as input_file:
                encryption_data = gpg.encrypt_file(
                    input_file,
                    recipients=key_fingerprint,
                    output=output_filename,
                    always_trust=True,
                    sign=signing_key,
                )

                # Check whether the encryption worked
                if not encryption_data.ok or encryption_data.status != "encryption ok":
                    self.logger.error(
                        f"Error encrypting file {file}: {encryption_data.status}"
                    )
                    # Print the stderr line too
                    self.logger.error(f"GPG STDERR: {encryption_data.stderr}")
                    raise exceptions.EncryptionError(
                        f"Error encrypting file {file}: {encryption_data.status}"
                    )

            encrypted_files[output_filename] = files[file]

        # Remove the temporary gnupg keychain files under f"{tmpdir}/.gnupg"
        if path.exists(f"{tmpdir}/.gnupg"):
            try:
                shutil.rmtree(f"{tmpdir}/.gnupg", ignore_errors=True)
            except FileNotFoundError as e:
                self.logger.warning(
                    f".gnupg deletion failed - FileNotFound but continuing: {e}"
                )

        return encrypted_files

    def decrypt_files(self, files: dict, private_key: str) -> dict:
        """Decrypt files using GPG.

        Args:
            files (dict): Dictionary of files to decrypt.
            private_key (str): Private key to use for decryption.

        Returns:
            dict: Dictionary of decrypted files.
        """
        # Use the local staging dir (or the path where the files live) as the gnupg home
        # Get the dirname of the first file in the list
        tmpdir = path.dirname(list(files.keys())[0])

        # Make the gnupg home directory
        makedirs(f"{tmpdir}/.gnupg", exist_ok=True)

        # Set up gnupg
        gpg = gnupg.GPG(gnupghome=f"{tmpdir}/.gnupg")

        private_key = private_key.replace("\\n", "\n")

        # Load the private key
        import_result = gpg.import_keys(private_key)
        # Check the key imported OK
        if not import_result.count or import_result.count == 0:
            self.logger.error("Error importing private key")
            raise exceptions.EncryptionError("Error importing private key")

        decrypted_files = {}
        for file in files:

            # If the filename contains .gpg or .pgp on the end, then the output file will be just that but without the extension,
            # if it doesn't contain .gpg or .pgp, then we'll add .decrypted to the end
            output_filename = (
                file.replace(".gpg", "").replace(".pgp", "")
                if file.endswith((".gpg", ".pgp"))
                else f"{file}.decrypted"
            )

            self.logger.info(f"Decrypting {file} to {output_filename}")

            with open(file, "rb") as input_file:
                decryption_data = gpg.decrypt_file(
                    input_file,
                    output=output_filename,
                )

                # Check whether the decryption worked
                if not decryption_data.ok or decryption_data.returncode != 0:
                    self.logger.error(
                        f"Error decrypting file {file}: {decryption_data.status}"
                    )
                    # Print the stderr line too
                    self.logger.error(f"GPG STDERR: {decryption_data.stderr}")

                    # Remove the temporary gnupg keychain files under f"{tmpdir}/.gnupg"
                    if path.exists(f"{tmpdir}/.gnupg"):
                        try:
                            shutil.rmtree(f"{tmpdir}/.gnupg", ignore_errors=True)
                        except FileNotFoundError as e:
                            self.logger.warning(
                                f".gnupg deletion failed - FileNotFound but continuing: {e}"
                            )

                    raise exceptions.DecryptionError(
                        f"Error decrypting file {file}: {decryption_data.status}"
                    )

            decrypted_files[output_filename] = files[file]

        # Remove the temporary gnupg keychain files under f"{tmpdir}/.gnupg"
        # Check if the directory exists first
        if path.exists(f"{tmpdir}/.gnupg"):
            try:
                shutil.rmtree(f"{tmpdir}/.gnupg", ignore_errors=True)
            except FileNotFoundError as e:
                self.logger.warning(
                    f".gnupg deletion failed - FileNotFound but continuing: {e}"
                )

        self.logger.debug(f"Returning decrypted files: {decrypted_files}")

        return decrypted_files

    def check_conditionals(self, remote_files: dict) -> dict:
        """Check conditionals of found files.

        Args:
            remote_files (dict): Dictionary of files found matching regex.

        Returns:
            dict: Dictionary of decrypted files.
        """
        for remote_file in list(remote_files):
            self.logger.info(f"Checking {remote_file}")

            # Check to see if there's a size condition
            meets_condition = True

            if "size" in self.source_file_spec["conditionals"]:
                self.logger.log(12, "Checking file size")
                min_size = self.source_file_spec["conditionals"]["size"].get("gt", None)
                max_size = self.source_file_spec["conditionals"]["size"].get("lt", None)

                file_size = remote_files[remote_file]["size"]

                if min_size and file_size <= min_size:
                    self.logger.info(
                        f"File is too small: Min size: [{min_size} B] Actual size:"
                        f" [{file_size} B]"
                    )
                    meets_condition = False

                if max_size and file_size >= max_size:
                    self.logger.info(
                        f"File is too big: Max size: [{max_size} B] Actual size:"
                        f" [{file_size} B]"
                    )
                    meets_condition = False

            if "age" in self.source_file_spec["conditionals"]:
                min_age = self.source_file_spec["conditionals"]["age"].get("gt", None)
                max_age = self.source_file_spec["conditionals"]["age"].get("lt", None)

                file_modified_time = remote_files[remote_file]["modified_time"]
                file_age = time.time() - file_modified_time

                self.logger.log(
                    12,
                    "Checking file age - Last modified time:"
                    f" {time.ctime(file_modified_time)} - Age in secs:"
                    f" {file_age} secs",
                )

                if min_age and file_age <= min_age:
                    self.logger.info(
                        f"File is too new: Min age: [{min_age} secs] Actual age:"
                        f" [{file_age} secs]"
                    )
                    meets_condition = False

                if max_age and file_age >= max_age:
                    self.logger.info(
                        f"File is too old: Max age: [{max_age} secs] Actual age:"
                        f" [{file_age} secs]"
                    )
                    meets_condition = False
            if not meets_condition:
                remote_files.pop(remote_file)
        if "count" in self.source_file_spec["conditionals"]:
            min_count = self.source_file_spec["conditionals"]["count"].get(
                "minCount", None
            )
            max_count = self.source_file_spec["conditionals"]["count"].get(
                "maxCount", None
            )
            # clear remote_files to ensure filewatch continues if count conditionals not correct
            if (max_count and (len(remote_files) > max_count)) or (
                min_count and (len(remote_files) < min_count)
            ):
                self.logger.info(
                    f"Number of found files {len(remote_files)} does not meet minCount/maxCount conditional: maxCount:{max_count if max_count else 'n/a'}, minCount:{min_count if min_count else 'n/a'}"
                )
                remote_files.clear()
        return remote_files



# pylint: skip-file
from tests.file_helper import list_test_files, write_test_file


def test_write_test_file_with_content(tmpdir):
    file_name = f"{tmpdir}/test.txt"
    content = "test1234"
    write_test_file(file_name, content)
    with open(file_name) as f:
        assert f.read() == content


def test_write_test_file_with_length(tmpdir):
    file_name = f"{tmpdir}/test.txt"
    length = 100
    write_test_file(file_name, length=length)
    with open(file_name) as f:
        assert len(f.read()) == length


def test_list_test_files(tmpdir):
    file_name = f"{tmpdir}/test.txt"
    content = "test1234"
    write_test_file(file_name, content)
    list = list_test_files(tmpdir, "test.txt", ",")
    assert list == file_name

    # Do the same but with a regex
    list = list_test_files(tmpdir, "test.*", ",")
    assert list == file_name



# pylint: skip-file
# ruff: noqa

import pytest

from opentaskpy.config.variablecaching import file
from opentaskpy.exceptions import CachingPluginError


def test_cacheable_variable_file(tmpdir):

    spec = {"task_id": "1234", "x": {"y": "value"}}

    kwargs = {"file": f"{tmpdir}/cacheable_variable.txt", "value": "newvalue"}

    file.run(**kwargs)

    # Check the new file has content of "newvalue"
    with open(f"{tmpdir}/cacheable_variable.txt", "r") as f:
        assert f.read() == "newvalue"


def test_cacheable_variable_file_invalid(tmpdir):

    spec = {"task_id": "1234", "x": {"y": "value"}}

    kwargs = {"file": f"/NONEXISTENT/NONEXISTENT", "value": "newvalue"}

    # Should throw a FileNotFoundError
    with pytest.raises(FileNotFoundError):
        file.run(**kwargs)


def test_cachable_variable_file_invalid_args():
    with pytest.raises(CachingPluginError):
        file.run()

    with pytest.raises(CachingPluginError):
        file.run(file="/tmp/file.txt")

    with pytest.raises(CachingPluginError):
        file.run(value="newvalue")



"""Utility functions for SSH."""

from logging import Logger

from paramiko import AutoAddPolicy, SSHClient


def setup_host_key_validation(client: SSHClient, spec: dict, logger: Logger) -> None:
    """Set up host key validation for an SSH client.

    Args:
        client (SSHClient): The SSH client to set up.
        spec (dict): The spec for the SSH connection.
        logger (logging.Logger): The logger to use.
    """
    logger.info("Loading system host keys")
    client.load_system_host_keys()

    if (
        "hostKeyValidation" in spec["protocol"]
        and spec["protocol"]["hostKeyValidation"]
    ):
        if "knownHostsFile" in spec["protocol"]:
            host_key = spec["protocol"]["knownHostsFile"]
            logger.info(f"Loading host keys from {host_key}")
            client.load_host_keys(host_key)
    else:
        client.set_missing_host_key_policy(AutoAddPolicy())
